import numpy as np
from typing import Callable, List

from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.layers import Dense, Input, BatchNormalization
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam

import matplotlib.pyplot as plt


class NNModel:
  """Neaural Networkモデルの構築"""
  def __init__(
    self,
    input_size: int,
    hidden_sizes: List,
    output_size: int,
    activation: str = "relu",
    output_activation: str = "relu",
    add_BN: bool = True
    ) -> None:
    """NN module を"""
    self.input = Input(shape=(input_size, ), name="input")
    self.output_size = Dense(output_size, name="output")
    self.add_BN = add_BN

    self.hidden_layers = []
    for idx, hidden_size in enumerate(hidden_sizes):
      if (idx+1) == len(self.hidden_layers):
        self.hidden_layers.append(Dense(hidden_size, output_activation, name=f"hidden_{idx+1}"))
        continue
      self.hidden_layers.append(Dense(hidden_size, activation, name=f"hidden_{idx+1}"))
  
  def build(self) -> Model:
    inputs = self.input
    for idx, hidden_layer in enumerate(self.hidden_layers):
      if idx == 0:
        x = hidden_layer(inputs)
        continue

      x = BatchNormalization()(x)
      x = hidden_layer(x)
    outputs = self.output_size(x)
    return Model(inputs, outputs)


def fit_nn(
  X,
  y,
  cv, 
  metrics: List,
  score_func: Callable[[np.ndarray, np.ndarray], float],
  NNModel: Model, 
  NN_MODEL_PARAMS: dict,
  n_epoch: int = 500, 
  loss: str = "mse",
  optimizer = "adam",
  verbose: int = -1,
  eary_stopping_rounds: int = 50,
  ):
  """NNの学習"""
  oof_pred = np.zeros(len(X), dtype=np.float32)
  models = []
  scores = []

  for i, (idx_train, idx_valid) in enumerate(cv):
    x_train, y_train = X[X.index.isin(idx_train)], y[idx_train]
    x_valid, y_valid = X[X.index.isin(idx_valid)], y[idx_valid]

    model = NNModel(**NN_MODEL_PARAMS)
    model = model.build()
    model.compile(
      optimizer,
      loss,
      metrics
    )
    history = model.fit(
          x_train,
          y_train,
          epochs=n_epoch,
          verbose=verbose,
          validation_data=(x_valid, y_valid),
          callbacks=[EarlyStopping(patience=eary_stopping_rounds)]
        )

    plot_history(history)

    oof = model.predict(x_valid)
    oof_pred[idx_valid] = oof.flatten()
    models.append(model)
    
    print("-"*50)
    if score_func.__name__ == "accuracy_score":
      print(f"score {i+1}:\t {score_func(y_valid, np.round(oof).astype(int))}")
    else:
      print(f"score {i+1}:\t {score_func(y_valid, oof)}")

  print("*"*50)
  if score_func.__name__ == "accuracy_score":
    score = score_func(y, np.round(oof_pred).astype(int))
  else:
    score = score_func(y, oof_pred)
  print(f"score {i+1}:\t {score}")

  return models, oof_pred, score


def plot_history(hist):
  # 損失値(Loss)の遷移のプロット
  plt.plot(hist.history['loss'],label="train set")
  plt.plot(hist.history['val_loss'],label="test set")
  plt.title('model loss')
  plt.xlabel('epoch')
  plt.ylabel('loss')
  plt.legend()
  plt.show()


import pandas as pd
import lightgbm as lgbm
from sklearn.model_selection import train_test_split

from tqdm.notebook import tqdm
from typing import List, Dict

class LGBMImputer:
  def __init__(
      self,
      feature_cols: List,
      target_cols: List,
      early_stopping_rounds: int = 20
    ) -> None:
    self.feature_cols = feature_cols
    self.target_cols = target_cols
    self.early_stopping_rounds = early_stopping_rounds
    self.models: Dict[str, lgbm.Booster] = {}

  def fit(self, df_input: pd.DataFrame):
    for target_col in tqdm(self.target_cols, desc="train lgbm models..."):
      # 欠損の行をデータセットとする
      train_ds, valid_ds = self._create_ds(df_input, target_col)
      params = self._feature_type_checker(train_ds.label)
      # 予測
      model = lgbm.train(
        {**params, "verbosity": -1},
        num_boost_round=500000,
        train_set=train_ds,
        callbacks=[
            lgbm.early_stopping(stopping_rounds=self.early_stopping_rounds, verbose=-1)
        ],
        valid_sets=[valid_ds]
      )
      self.models[target_col] = model
      
    return self.transform(df_input)


  def transform(self, df_input: pd.DataFrame):
    df_input_cp = df_input.copy(deep=True)

    for target_col in tqdm(self.target_cols, desc="predict from lgbm models..."):
      if df_input[target_col].isnull().sum() > 0:
        # 欠損の行をデータセットとする
        _, valid_ds = self._create_ds(df_input, target_col)
        df_input_cp.loc[valid_ds.data.index, target_col] = self.models[target_col].predict(valid_ds.data, num_iteration=self.models[target_col].best_iteration)
    return df_input_cp


  def _create_ds(self, df_input: pd.DataFrame, target_col: str):
    # 欠損の行をデータセットとする / 欠損補完するカラムはテストデータとする
    idx_miss = df_input[df_input[target_col].isnull()].index
    if len(idx_miss) != 0:
      train_ds = lgbm.Dataset(
        df_input[~df_input.index.isin(idx_miss)][self.feature_cols],
        df_input[~df_input.index.isin(idx_miss)][target_col],
      )
      valid_ds = lgbm.Dataset(
        df_input[df_input.index.isin(idx_miss)][self.feature_cols],
        df_input[df_input.index.isin(idx_miss)][target_col],
      )
    else:
      # 訓練データに欠損値がないとき
      df_train, df_valid = train_test_split(df_input, test_size=.2, random_state=712)
      train_ds = lgbm.Dataset(
        df_input[df_input.index.isin(df_train.index)][self.feature_cols],
        df_input[df_input.index.isin(df_train.index)][target_col],
      )
      valid_ds = lgbm.Dataset(
        df_input[df_input.index.isin(df_valid.index)][self.feature_cols],
        df_input[df_input.index.isin(df_valid.index)][target_col],
      )
    return train_ds, valid_ds

  def _feature_type_checker(self, series: pd.Series) -> dict:
    # 自動で型を判定してパラメータを返す
    if pd.api.types.is_numeric_dtype(series):
      # 連続変数のとき
      return {
        'objective': 'regression'
      }
    else:
      # カテゴリカル変数のとき
      nuni = series.dropna().nunique()
      if nuni == 2:
        return {
          'objective': 'binary'
        }
      elif nuni > 2:
        return {
          'objective': 'multiclass',
          'num_class': nuni + 1
        }


import pandas as pd
import numpy as np
from typing import Callable
import xgboost as xgb

from ml_vite.util.timer import Timer

def fit_xgb(
    X: pd.DataFrame, 
    y, 
    cv, 
    params, 
    score_func: Callable[[np.ndarray, np.ndarray], float],
    early_stopping_rounds: int = 20,
  ):
  oof_pred = np.zeros(len(X), dtype=np.float32)
  models = []
  scores = []

  for i, (idx_train, idx_valid) in enumerate(cv):
    x_train, y_train = X[X.index.isin(idx_train)], y[idx_train]
    x_valid, y_valid = X[X.index.isin(idx_valid)], y[idx_valid]

    ds_train = xgb.DMatrix(x_train, y_train)
    ds_valid = xgb.DMatrix(x_valid, y_valid)

    with Timer(prefix=f"fit ========== Fold: {i + 1}"):
      model = xgb.train(
        params,
        dtrain=ds_train,
        evals=[(ds_train, "train"), (ds_valid, "valid")],
        early_stopping_rounds=early_stopping_rounds,
        num_boost_round=500000,
        verbose_eval=200
      )

      pred_i = model.predict(ds_valid, ntree_limit=model.best_ntree_limit)
      oof_pred[idx_valid] = pred_i

      if score_func.__name__ == "accuracy_score":
        print(f" - fold{i + 1} - :\t {score_func(y_valid, np.round(oof_pred).astype(int))}")
      else:
        print(f" - fold{i + 1} - :\t {score_func(y_valid, oof_pred)}")
      print(f" - fold{i + 1} - {score:.4f}")

      scores.append(score)
      models.append(model)
    
  if score_func.__name__ == "accuracy_score":
    score = score_func(y, np.round(oof_pred).astype(int))
  else:
    score = score_func(y, oof_pred)
    
  print("=" * 50)
  print(f"FINISH: Whole Score: {score:.4f}")

  return oof_pred, models, score


from ml_vite.feature_eng.base_block import AbstractBaseBlock, run_blocks
from ml_vite.feature_eng.encoding_block import TargetEncodingBlock, LabelEncodingBlock, CountEncodingBlock, OneHotEncodingBlock


from typing import List
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from ml_vite.util.types import t_cv
from ml_vite.util.error import CustomErrorMsg
from ml_vite.feature_eng.base_block import AbstractBaseBlock

class TargetEncodingBlock(AbstractBaseBlock):
  """Target Encoding"""
  def __init__(self, target_col, agg: List, cv: t_cv, **kwrgs) -> None:
    super().__init__(**kwrgs)
    self.target = target_col
    self.agg = agg
    self.cv = cv

  def fit(self, df_input: pd.DataFrame, y):
    self.dics_ = []
    if y not in df_input.columns:
      raise CustomErrorMsg(f"df_input に目的変数 {y} がありません")

    df_out = pd.DataFrame(index=df_input.index)

    for idx_feature, idx_valid in self.cv:
      # train と valid を作成
      df_feature = df_input[df_input.index.isin(idx_feature)]
      df_valid = df_input[df_input.index.isin(idx_valid)]
      _dic = df_feature.groupby(self.target)[y].agg(self.agg).to_dict()
      self.dics_.append(_dic)
      # agg で回す
      for agg in self.agg:
        df_out.loc[idx_valid, f"target={self.target}_agg_func={agg}"] = df_valid[self.target].map(_dic[agg])
    return df_out

  def transform(self, df_input: pd.DataFrame):
    df_out = pd.DataFrame()
    for agg in self.agg:
      _df = pd.DataFrame()
      for i, _dic in enumerate(self.dics_):
        _df[i] = df_input[self.target].map(_dic[agg])
      df_out[f"target={self.target}_agg_func={agg}"] = _df.mean(axis=1)
    return df_out


class OneHotEncodingBlock(AbstractBaseBlock):
    """ダミー変数の作成"""
    def __init__(self, target_col: str, drop_first: bool, **kwrgs) -> None:
        super().__init__(**kwrgs)
        self.target_col = target_col
        self.drop_first = drop_first
    
    def fit(self, df_input: pd.DataFrame, y=None):
        df_train_out = pd.get_dummies(df_input[self.target_col], drop_first=self.drop_first).add_prefix(f"{self.target_col}_")
        self.train_column_ = df_train_out.columns
        return df_train_out

    def transform(self, df_input: pd.DataFrame):
        return pd.get_dummies(df_input[self.target_col], drop_first=self.drop_first).add_prefix(f"{self.target_col}_").reindex(columns=self.train_column_, fill_value=0)


class CountEncodingBlock(AbstractBaseBlock):
  """CountEncoding"""
  def __init__(self, target_col: str, **kwrgs) -> None:
    super().__init__(**kwrgs)
    self.target_col = target_col

  def fit(self, df_input: pd.DataFrame, y=None):
    self.vc_ = df_input[self.target_col].value_counts()
    return self.transform(df_input)

  def transform(self, df_input: pd.DataFrame):
    return pd.DataFrame(df_input[self.target_col].map(self.vc_))


class LabelEncodingBlock(AbstractBaseBlock):
  """LabelEncoding"""
  def __init__(self, target_col: str) -> None:
    super().__init__()
    self.le = LabelEncoder()
    self.target_col = target_col

  def fit(self, df_input: pd.DataFrame, y=None):
    self.le.fit(df_input[self.target_col])
    return self.transform(df_input)

  def transform(self, df_input: pd.DataFrame):
    return pd.DataFrame({
        self.target_col: self.le.fit_transform(df_input[self.target_col])
    })


import pandas as pd
from ml_vite.util.timer import Timer

class AbstractBaseBlock:
  """BaseBlock 定義"""
  def __init__(self, parent_blocks = None):
    self.parent_blocks = [] if parent_blocks is None else parent_blocks

  def fit(self, input_df: pd.DataFrame, y=None):        
    return self.transform(input_df)

  def transform(self, input_df: pd.DataFrame) -> pd.DataFrame:
    raise NotImplementedError()


def run_blocks(input_df, blocks, y=None, test=False):
  """BaseBlock 実行関数"""
  df_out = pd.DataFrame()

  for block in blocks:
    if block.parent_blocks:
      _df = run_blocks(input_df=input_df, blocks=block.parent_blocks, y=y, test=test)
    else:
      _df = input_df

    with Timer(prefix='\t- {}'.format(str(block))):
      if not test:
        out_i = block.fit(_df, y=y)
      else:
        out_i = block.transform(_df)

    assert len(input_df) == len(out_i), block
    name = block.__class__.__name__
    df_out = pd.concat([df_out, out_i.add_suffix(f"@{name}")], axis=1)

  return df_out

def help():
  # TODO ヘルプの実装
  print(AbstractBaseBlock.__name__)



import datetime
import pandas as pd

def save_sub(
  df_sub: pd.DataFrame,
  nb_num: str,
  score: float,
  save_path: str
  ) -> None:
  """submission の作成"""
  now = datetime.datetime.now()
  time = now.strftime('%m_%d_%H_%M')

  df_sub.to_csv(f"{save_path}/{nb_num}_sub_{time}_{round(score, 4)}.csv", index=False)



from time import time

class Timer:
  def __init__(self, logger=None, format_str="{:.3f}[s]", prefix=None, suffix=None, sep=" "):
    if prefix:
      format_str = str(prefix) + sep + format_str
    if suffix:
      format_str = format_str + sep + str(suffix)
    self.format_str = format_str
    self.logger = logger
    self.start = None
    self.end = None

  @property
  def duration(self):
    if self.end is None:
      return 0
    return self.end - self.start

  def __enter__(self):
    self.start = time()

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.end = time()
    out_str = self.format_str.format(self.duration)
    if self.logger:
      self.logger.info(out_str)
    else:
      print(out_str)



import pandas as pd
import numpy as np
import lightgbm as lgbm
import matplotlib.pyplot as plt
import seaborn as sns
import japanize_matplotlib

from typing import Callable

from ml_vite.util.timer import Timer


def fit_lgbm(
    X: pd.DataFrame,
    y,
    cv,
    params,
    score_func: Callable[[np.ndarray, np.ndarray], float],
  ):
  """LightGBMでfitする"""
  oof_pred = np.zeros(len(X), dtype=np.float32)
  models = []
  scores = []

  for i, (idx_train, idx_valid) in enumerate(cv):
    x_train, y_train = X[X.index.isin(idx_train)], y[idx_train]
    x_valid, y_valid = X[X.index.isin(idx_valid)], y[idx_valid]

    ds_train = lgbm.Dataset(x_train, y_train)
    ds_valid = lgbm.Dataset(x_valid, y_valid)

    with Timer(prefix=f"fit ========== Fold: {i + 1}"):
      model = lgbm.train(
        params,
        ds_train,
        valid_names=["train, valid"],
        valid_sets=[ds_train, ds_valid]
      )

      pred_i = model.predict(x_valid, num_iteration=model.best_iteration)
      oof_pred[idx_valid] = pred_i

      if score_func.__name__ == "accuracy_score":
        score = score_func(y_valid, np.round(pred_i).astype(int))
        print(f" - fold{i + 1} - :\t {score}")
      else:
        score = score_func(y_valid, pred_i)
        print(f" - fold{i + 1} - :\t {score}")

      scores.append(score)
      models.append(model)
      
  if score_func.__name__ == "accuracy_score":
    score = score_func(y, np.round(oof_pred).astype(int))
  else:
    score = score_func(y, oof_pred)
    
  print("=" * 50)
  print(f"FINISH: Whole Score: {score:.4f}")

  return oof_pred, models, score


def visualize_importance(models, feat_train_df, top_n):
  """lightGBM の model 配列の feature importance を plot する
  CVごとのブレを boxen plot として表現

  args:
      models:
          List of lightGBM models
      feat_train_df:
          学習時に使った DataFrame
  """
  feature_importance_df = pd.DataFrame()
  for i, model in enumerate(models):
    _df = pd.DataFrame()
    _df["feature_importance"] = model.feature_importance(importance_type='split')
    _df["column"] = feat_train_df.columns
    _df["fold"] = i + 1
    feature_importance_df = pd.concat([feature_importance_df, _df], 
                                        axis=0, ignore_index=True)

  order = feature_importance_df.groupby("column")\
    .sum()[["feature_importance"]]\
    .sort_values("feature_importance", ascending=False).index[:top_n]

  fig, ax = plt.subplots(figsize=(8, max(6, len(order) * .25)))
  sns.boxenplot(data=feature_importance_df, 
                x="feature_importance", 
                y="column", 
                order=order, 
                ax=ax, 
                palette="viridis", 
                orient="h")
  ax.tick_params(axis="x", rotation=90)
  ax.set_title("Importance")
  ax.grid()
  fig.tight_layout()
  plt.show()
  return fig, ax


from setuptools import setup, find_packages
from os import path

def get_version():
    version_filepath = path.join(path.dirname(__file__), 'ml_vite', 'version.py')
    with open(version_filepath) as f:
        for line in f:
            if line.startswith('__version__'):
                return line.strip().split()[-1][1:-1]

install_requires = [
  "pandas",
  "numpy",
  "tqdm"
]

setup(
  name="ml_vite",
  version=get_version(),
  install_requires=install_requires,
  author="kazuyan",
  packages=find_packages()
)